from __future__ import annotations

import logging
import typing as t
from copy import copy
from uuid import uuid4

from datasets import Dataset

from ragas.embeddings import LlamaIndexEmbeddingsWrapper
from ragas.evaluation import evaluate as ragas_evaluate
from ragas.exceptions import ExceptionInRunner
from ragas.executor import Executor
from ragas.llms import LlamaIndexLLMWrapper
from ragas.run_config import RunConfig
from ragas.validation import EVALMODE_TO_COLUMNS, validate_evaluation_modes

if t.TYPE_CHECKING:
    from llama_index.core.base.embeddings.base import (
        BaseEmbedding as LlamaIndexEmbeddings,
    )
    from llama_index.core.base.llms.base import BaseLLM as LlamaindexLLM

    from ragas.evaluation import Result
    from ragas.metrics.base import Metric


logger = logging.getLogger(__name__)


def validate_dataset(dataset: dict, metrics: list[Metric]):
    # change EVALMODE_TO_COLUMNS for usecase with no contexts and answer
    evalmod_to_columns_llamaindex = copy(EVALMODE_TO_COLUMNS)
    for mode in EVALMODE_TO_COLUMNS:
        if "answer" in EVALMODE_TO_COLUMNS[mode]:
            EVALMODE_TO_COLUMNS[mode].remove("answer")
        if "contexts" in EVALMODE_TO_COLUMNS[mode]:
            EVALMODE_TO_COLUMNS[mode].remove("contexts")

    hf_dataset = Dataset.from_dict(dataset)
    validate_evaluation_modes(hf_dataset, metrics, evalmod_to_columns_llamaindex)


def evaluate(
    query_engine,
    dataset: Dataset,
    metrics: list[Metric],
    llm: t.Optional[LlamaindexLLM] = None,
    embeddings: t.Optional[LlamaIndexEmbeddings] = None,
    raise_exceptions: bool = False,
    column_map: t.Optional[t.Dict[str, str]] = None,
    run_config: t.Optional[RunConfig] = None,
) -> Result:
    column_map = column_map or {}

    # wrap llms and embeddings
    li_llm = None
    if llm is not None:
        li_llm = LlamaIndexLLMWrapper(llm)
    li_embeddings = None
    if embeddings is not None:
        li_embeddings = LlamaIndexEmbeddingsWrapper(embeddings)

    # validate and transform dataset
    if dataset is None:
        raise ValueError("Provide dataset!")

    exec = Executor(
        desc="Running Query Engine",
        keep_progress_bar=True,
        raise_exceptions=raise_exceptions,
        run_config=run_config,
    )

    # get query
    queries = dataset["question"]
    for i, q in enumerate(queries):
        exec.submit(query_engine.aquery, q, name=f"query-{i}")

    answers: t.List[str] = []
    contexts: t.List[t.List[str]] = []
    try:
        results = exec.results()
        if results == []:
            raise ExceptionInRunner()
    except Exception as e:
        raise e
    else:
        for r in results:
            answers.append(r.response)
            contexts.append([n.node.text for n in r.source_nodes])

    # create HF dataset
    hf_dataset = Dataset.from_dict(
        {
            "question": queries,
            "contexts": contexts,
            "answer": answers,
        }
    )
    if "ground_truth" in dataset.column_names:
        hf_dataset = hf_dataset.add_column(
            name="ground_truth",
            column=dataset["ground_truth"],
            new_fingerprint=str(uuid4()),
        )

    results = ragas_evaluate(
        dataset=hf_dataset,
        metrics=metrics,
        llm=li_llm,
        embeddings=li_embeddings,
        raise_exceptions=raise_exceptions,
    )

    return results



"""
Q - question
A - answer: generated_text from RAG pipeline
C - contexts: context used for generation
G - ground_truth: ground truth answer
"""

from __future__ import annotations

import asyncio
import logging
import typing as t
from abc import ABC, abstractmethod
from collections import Counter
from dataclasses import dataclass
from enum import Enum

from ragas.callbacks import new_group
from ragas.run_config import RunConfig

if t.TYPE_CHECKING:
    from langchain_core.callbacks import Callbacks

    from ragas.embeddings import BaseRagasEmbeddings
    from ragas.llms import BaseRagasLLM

from pysbd import Segmenter
from pysbd.languages import LANGUAGE_CODES

logger = logging.getLogger(__name__)


LANGUAGE_CODES = {v.__name__.lower(): k for k, v in LANGUAGE_CODES.items()}

EvaluationMode = Enum("EvaluationMode", "qac qa qc gc ga qga qcg ca")


def get_required_columns(
    eval_mod: EvaluationMode, ignore_columns: t.Optional[t.List[str]] = None
) -> t.List[str]:
    if eval_mod == EvaluationMode.qac:
        keys = ["question", "answer", "contexts"]
    elif eval_mod == EvaluationMode.qa:
        keys = ["question", "answer"]
    elif eval_mod == EvaluationMode.qc:
        keys = ["question", "contexts"]
    elif eval_mod == EvaluationMode.gc:
        keys = ["contexts", "ground_truth"]
    elif eval_mod == EvaluationMode.ga:
        keys = ["answer", "ground_truth"]
    elif eval_mod == EvaluationMode.qga:
        keys = ["question", "contexts", "answer", "ground_truth"]
    elif eval_mod == EvaluationMode.qcg:
        keys = ["question", "contexts", "ground_truth"]
    elif eval_mod == EvaluationMode.ca:
        keys = ["contexts", "answer"]
    ignore_columns = ignore_columns or []

    return [k for k in keys if k not in ignore_columns]


@dataclass
class Metric(ABC):
    @property
    @abstractmethod
    def name(self) -> str:
        ...

    @property
    @abstractmethod
    def evaluation_mode(self) -> EvaluationMode:
        ...

    @abstractmethod
    def init(self, run_config: RunConfig):
        """
        This method will lazy initialize the model.
        """
        ...

    def adapt(self, language: str, cache_dir: t.Optional[str] = None) -> None:
        """
        Adapt the metric to a different language.
        """
        raise NotImplementedError(
            "adapt() is not implemented for {} metric".format(self.name)
        )

    def save(self, cache_dir: t.Optional[str] = None) -> None:
        """
        Save the metric to a path.
        """
        raise NotImplementedError(
            "adapt() is not implemented for {} metric".format(self.name)
        )

    def score(self: t.Self, row: t.Dict, callbacks: Callbacks = None) -> float:
        callbacks = callbacks or []
        rm, group_cm = new_group(self.name, inputs=row, callbacks=callbacks)
        try:
            loop = asyncio.get_event_loop()
            score = loop.run_until_complete(self._ascore(row=row, callbacks=group_cm))
        except Exception as e:
            if not group_cm.ended:
                rm.on_chain_error(e)
            raise e
        else:
            if not group_cm.ended:
                rm.on_chain_end({"output": score})
        return score

    async def ascore(
        self: t.Self,
        row: t.Dict,
        callbacks: Callbacks = None,
        timeout: t.Optional[float] = None,
    ) -> float:
        callbacks = callbacks or []
        rm, group_cm = new_group(self.name, inputs=row, callbacks=callbacks)
        try:
            score = await asyncio.wait_for(
                self._ascore(row=row, callbacks=group_cm),
                timeout=timeout,
            )
        except Exception as e:
            if not group_cm.ended:
                rm.on_chain_error(e)
            raise e
        else:
            if not group_cm.ended:
                rm.on_chain_end({"output": score})
        return score

    @abstractmethod
    async def _ascore(self, row: t.Dict, callbacks: Callbacks) -> float:
        ...


@dataclass
class MetricWithLLM(Metric):
    llm: t.Optional[BaseRagasLLM] = None

    def init(self, run_config: RunConfig):
        """
        Init any models in the metric, this is invoked before evaluate()
        to load all the models
        Also check if the api key is valid for OpenAI and AzureOpenAI
        """
        if self.llm is None:
            raise ValueError(
                f"Metric '{self.name}' has no valid LLM provided (self.llm is None). Please initantiate a the metric with an LLM to run."  # noqa
            )
        self.llm.set_run_config(run_config)


@dataclass
class MetricWithEmbeddings(Metric):
    embeddings: t.Optional[BaseRagasEmbeddings] = None

    def init(self, run_config: RunConfig):
        """
        Init any models in the metric, this is invoked before evaluate()
        to load all the models
        Also check if the api key is valid for OpenAI and AzureOpenAI
        """
        if self.embeddings is None:
            raise ValueError(
                f"Metric '{self.name}' has no valid embeddings provided (self.embeddings is None). Please initantiate a the metric with an embeddings to run."  # noqa
            )
        self.embeddings.set_run_config(run_config)


class Ensember:
    """
    Combine multiple llm outputs for same input (n>1) to a single output
    """

    def from_discrete(self, inputs: list[list[t.Dict]], attribute: str):
        """
        Simple majority voting for binary values, ie [0,0,1] -> 0
        inputs: list of list of dicts each containing verdict for a single input
        """

        if not isinstance(inputs, list):
            inputs = [inputs]

        if not all(len(item) == len(inputs[0]) for item in inputs):
            logger.warning("All inputs must have the same length")
            return inputs[0]

        if not all(attribute in item for input in inputs for item in input):
            logger.warning(f"All inputs must have {attribute} attribute")
            return inputs[0]

        if len(inputs) == 1:
            return inputs[0]

        verdict_agg = []
        for i in range(len(inputs[0])):
            item = inputs[0][i]
            verdicts = [inputs[k][i][attribute] for k in range(len(inputs))]
            verdict_counts = dict(Counter(verdicts).most_common())
            item[attribute] = list(verdict_counts.keys())[0]
            verdict_agg.append(item)

        return verdict_agg


def get_segmenter(
    language: str = "english", clean: bool = False, char_span: bool = False
):
    """
    Get a sentence segmenter for a given language
    """
    language = language.lower()
    if language not in LANGUAGE_CODES:
        raise ValueError(
            f"Language '{language}' not supported. Supported languages: {LANGUAGE_CODES.keys()}"
        )
    return Segmenter(
        language=LANGUAGE_CODES[language], clean=clean, char_span=char_span
    )


def is_reproducable(metric: Metric) -> bool:
    return hasattr(metric, "_reproducibility")


ensembler = Ensember()



import pytest
from langchain_core.outputs import Generation, LLMResult
from ragas_experimental.llms.prompt import StringIO, StringPrompt

from ragas.llms.base import BaseRagasLLM
from ragas.llms.prompt import PromptValue
from ragas.run_config import RunConfig


class EchoLLM(BaseRagasLLM):
    def generate_text(  # type: ignore
        self,
        prompt: PromptValue,
    ) -> LLMResult:
        return LLMResult(generations=[[Generation(text=prompt.to_string())]])

    async def agenerate_text(  # type: ignore
        self,
        prompt: PromptValue,
    ) -> LLMResult:
        return LLMResult(generations=[[Generation(text=prompt.to_string())]])


@pytest.mark.asyncio
async def test_string_prompt():
    echo_llm = EchoLLM(run_config=RunConfig())
    prompt = StringPrompt(llm=echo_llm)
    assert await prompt.generate("hello") == "hello"


expected_generate_output_signature = """\
Please return the output in the following JSON format based on the StringIO model:
{
    "text": "str"
}\
"""


def test_process_fields():
    from enum import Enum

    from pydantic import BaseModel
    from ragas_experimental.llms.prompt import PydanticPrompt, StringIO

    class Categories(str, Enum):
        science = "science"
        commerce = "commerce"
        agriculture = "agriculture"
        economics = "economics"

    class InputModel(BaseModel):
        category: Categories

    class JokeGenerator(PydanticPrompt[InputModel, StringIO]):
        instruction = "Generate a joke in the category of {category}."
        output_model = StringIO

    echo_llm = EchoLLM(run_config=RunConfig())
    p = JokeGenerator(llm=echo_llm)
    _ = p.generate_output_signature()

    # assert expected_generate_output_signature == generation


@pytest.mark.asyncio
async def test_pydantic_prompt_io():
    from ragas_experimental.llms.prompt import PydanticPrompt, StringIO

    class Prompt(PydanticPrompt[StringIO, StringIO]):
        instruction = ""
        input_model = StringIO
        output_model = StringIO

    llm = EchoLLM(run_config=RunConfig())
    p = Prompt(llm=llm)
    assert p.input_model == StringIO
    assert p.output_model == StringIO

    assert p.generate_examples() == ""


def test_pydantic_prompt_examples():
    from ragas_experimental.llms.prompt import PydanticPrompt

    class Prompt(PydanticPrompt[StringIO, StringIO]):
        instruction = ""
        input_model = StringIO
        output_model = StringIO
        examples = [
            (StringIO(text="hello"), StringIO(text="hello")),
            (StringIO(text="world"), StringIO(text="world")),
        ]

    llm = EchoLLM(run_config=RunConfig())
    _ = Prompt(llm=llm)
    # assert p.generate_examples() == "hello -> hello\nworld -> world"



import json
import logging
import typing as t
from dataclasses import dataclass, field

import numpy as np
from langchain_core.documents import Document as LCDocument
from ragas_experimental.testset.graph import Node
from ragas_experimental.testset.questions.base import (
    DEFAULT_DISTRIBUTION,
    QAC,
    QAGenerator,
    QuestionLength,
    QuestionStyle,
)
from ragas_experimental.testset.questions.prompts import (
    critic_question,
    order_sections_by_relevance,
    question_answering,
    specific_question_from_keyphrase,
)
from ragas_experimental.testset.questions.queries import CHILD_NODES_QUERY
from ragas_experimental.testset.utils import rng

from ragas.executor import Executor
from ragas.llms.prompt import Prompt

logger = logging.getLogger(__name__)


@dataclass
class SpecificQA(QAGenerator):
    name: str = "SpecificQuestion"
    generate_question_prompt: Prompt = field(
        default_factory=lambda: specific_question_from_keyphrase
    )
    generate_answer_prompt: Prompt = field(default_factory=lambda: question_answering)
    critic_question_prompt: Prompt = field(default_factory=lambda: critic_question)
    order_sections_prompt: Prompt = field(
        default_factory=lambda: order_sections_by_relevance
    )

    """
    specic questions from particular sections of particular document
    for example, what are model architecture details used in OpenMoE paper?
    """

    async def generate_questions(
        self, query, kwargs, distribution=DEFAULT_DISTRIBUTION, num_samples=5
    ):
        assert self.llm is not None, "LLM is not initialized"
        query = query or CHILD_NODES_QUERY
        kwargs = kwargs or {}
        nodes = self.query_nodes(query, kwargs)
        num_nodes = min(num_samples, len(nodes))
        nodes = rng.choice(nodes, size=num_nodes, replace=False)
        seed_per_results = num_samples // len(nodes)
        reminder = num_samples - seed_per_results * num_nodes
        seeds = [seed_per_results] * num_nodes
        seeds[-1] += reminder

        nodes_and_keyphraes = []
        for node, seed in zip(nodes, seeds):
            seperators = [
                rel.properties["seperator"]
                for rel in node.relationships
                if rel.label == "child"
            ]
            if len(seperators) > 1:
                p_vlaue = self.order_sections_prompt.format(sections=seperators)
                output = await self.llm.generate(p_vlaue)
                output = json.loads(output.generations[0][0].text)
                # TODO: make sure that prompt does not modify the seperator. Ideally ouput ordering by index
                headings_array = (
                    np.array(output.get("high"))
                    if output.get("high")
                    else np.array(list(output.values()))
                )
                selected_heading = rng.choice(headings_array, size=seed)
            else:
                # TODO: inspect and handle better
                selected_heading = seperators[0:1]

            target_nodes = [
                relation.target
                for relation in node.relationships
                if relation.label == "child"
                and relation.source.id == node.id
                and any(
                    text in relation.properties["seperator"]
                    for text in selected_heading
                )
            ]

            keyphrases = [
                node.properties["metadata"]["keyphrases"] for node in target_nodes
            ]
            keyphrases = list(
                set([phrase for phrases in keyphrases for phrase in phrases])
            )
            keyphrases = rng.choice(np.array(keyphrases), size=seed)
            nodes_and_keyphraes.extend(
                [(target_nodes, keyphrase) for keyphrase in keyphrases]
            )

        exec = Executor(
            desc="Generating",
            keep_progress_bar=True,
            raise_exceptions=True,
            run_config=None,
        )

        index = 0
        for dist, prob in distribution.items():
            style, length = dist
            for i in range(int(prob * num_samples)):
                exec.submit(
                    self.generate_question,
                    nodes_and_keyphraes[index][0],
                    style,
                    length,
                    {"keyphrase": nodes_and_keyphraes[index][1]},
                )
                index += 1

        remaining_size = num_samples - index
        if remaining_size != 0:
            choices = np.array(distribution.keys())
            prob = np.array(distribution.values())
            random_distribution = rng.choice(choices, p=prob, size=remaining_size)
            for dist in random_distribution:
                style, length = dist
                exec.submit(
                    self.generate_question,
                    nodes_and_keyphraes[index][0],
                    style,
                    length,
                    {"keyphrase": nodes_and_keyphraes[index][1]},
                )
                index += 1

        return exec.results()

    async def generate_question(
        self,
        nodes: t.List[Node],
        style: QuestionStyle,
        length: QuestionLength,
        kwargs: t.Optional[dict] = None,
    ) -> QAC:
        assert self.llm is not None, "LLM is not initialized"
        assert self.embedding is not None, "Embedding is not initialized"
        kwargs = kwargs or {}

        keyphrase = kwargs["keyphrase"]
        title = nodes[0].properties["metadata"]["title"]
        text = "\n\n".join([node.properties["page_content"] for node in nodes])
        try:
            source = self.retrieve_chunks(nodes)
            if source:
                p_value = self.generate_question_prompt.format(
                    title=title, keyphrase=keyphrase, text=text
                )
                question = await self.llm.generate(p_value)
                question = question.generations[0][0].text

                critic_verdict = await self.critic_question(question)
                if critic_verdict:
                    question = await self.modify_question(question, style, length)
                    answer = await self.generate_answer(question, source)
                    return QAC(
                        question=question,
                        answer=answer,
                        source=source,
                        name=self.name,
                        style=style,
                        length=length,
                    )
                else:
                    logger.warning("Critic rejected the question: %s", question)
                    return QAC()
            else:
                logger.warning("Failed to retrieve chunks for nodes: %s", nodes)
                return QAC()

        except Exception as e:
            logging.warning("Failed to generate question: %s", e)
            raise e

    async def critic_question(self, question: str) -> bool:
        assert self.llm is not None, "LLM is not initialized"
        output = await self.llm.generate(critic_question.format(question=question))
        output = json.loads(output.generations[0][0].text)
        return all(score >= 2 for score in output.values())

    def retrieve_chunks(self, nodes: t.List[Node], kwargs: dict | None = None) -> t.Any:
        assert self.llm is not None, "LLM is not initialized"
        documents = [
            LCDocument(
                page_content=node.properties["page_content"],
                metadata=node.properties["metadata"],
            )
            for node in nodes
        ]
        return documents

    async def generate_answer(self, question: str, chunks: t.List[LCDocument]) -> t.Any:
        assert self.llm is not None, "LLM is not initialized"
        text = "\n\n".join([chunk.page_content for chunk in chunks])
        output = await self.llm.generate(
            self.generate_answer_prompt.format(question=question, text=text)
        )
        return output.generations[0][0].text



import pytest
from langchain_core.messages import AIMessage
from langchain_core.outputs import ChatGeneration, LLMResult

from ragas.cost import (
    CostCallbackHandler,
    TokenUsage,
    get_token_usage_for_anthropic,
    get_token_usage_for_openai,
)

"""
TODO: things to test
- get usage from LLM Result
- estimate cost works for different API providers 
- openai with multiple n
- anthropic
- anthropic with multiple n
"""


def test_token_usage():
    x = TokenUsage(input_tokens=10, output_tokens=20)
    y = TokenUsage(input_tokens=5, output_tokens=15)
    assert (x + y).input_tokens == 15
    assert (x + y).output_tokens == 35

    with pytest.raises(ValueError):
        x.model = "openai"
        y.model = "gpt3"
        _ = x + y

    # test equals
    assert x == x
    assert y != x
    z = TokenUsage(input_tokens=10, output_tokens=20)
    z_with_model = TokenUsage(input_tokens=10, output_tokens=20, model="openai")
    z_same_with_model = TokenUsage(input_tokens=10, output_tokens=20, model="openai")
    assert z_with_model != z
    assert z_same_with_model == z_with_model

    # test same model
    assert z_with_model.is_same_model(z_same_with_model)
    assert not z_with_model.is_same_model(z)


def test_token_usage_cost():
    x = TokenUsage(input_tokens=10, output_tokens=20)
    assert x.cost(cost_per_input_token=0.1, cost_per_output_token=0.2) == 5.0


openai_llm_result = LLMResult(
    generations=[[ChatGeneration(message=AIMessage(content="Hello, world!"))]],
    llm_output={
        "token_usage": {
            "completion_tokens": 10,
            "prompt_tokens": 10,
            "total_tokens": 20,
        },
        "model_name": "gpt-4o",
        "system_fingerprint": "fp_2eie",
    },
)

athropic_llm_result = LLMResult(
    generations=[
        [
            ChatGeneration(
                message=AIMessage(
                    content="Hello, world!",
                    response_metadata={
                        "id": "msg_01UHjFfUr",
                        "model": "claude-3-opus-20240229",
                        "stop_reason": "end_turn",
                        "stop_sequence": None,
                        "usage": {"input_tokens": 9, "output_tokens": 12},
                    },
                )
            )
        ]
    ],
    llm_output={},
)


def test_parse_llm_results():
    # openai
    token_usage = get_token_usage_for_openai(openai_llm_result)
    assert token_usage == TokenUsage(input_tokens=10, output_tokens=10)

    # anthropic
    token_usage = get_token_usage_for_anthropic(athropic_llm_result)
    assert token_usage == TokenUsage(input_tokens=9, output_tokens=12)


def test_cost_callback_handler():
    cost_cb = CostCallbackHandler(token_usage_parser=get_token_usage_for_openai)
    cost_cb.on_llm_end(openai_llm_result)

    # cost
    assert cost_cb.total_tokens() == TokenUsage(input_tokens=10, output_tokens=10)

    assert cost_cb.total_cost(0.1) == 2.0
    assert (
        cost_cb.total_cost(cost_per_input_token=0.1, cost_per_output_token=0.1) == 2.0
    )



from __future__ import annotations

import logging

from datasets import Dataset, Sequence

from ragas.metrics._context_precision import ContextPrecision
from ragas.metrics.base import EvaluationMode, Metric

logger = logging.getLogger(__name__)


def remap_column_names(dataset: Dataset, column_map: dict[str, str]) -> Dataset:
    """
    Remap the column names in case dataset uses different column names
    """

    inverse_column_map = {v: k for k, v in column_map.items()}
    return dataset.rename_columns(inverse_column_map)


def handle_deprecated_ground_truths(ds: Dataset) -> Dataset:
    if "ground_truths" in ds.features and "ground_truth" not in ds.features:
        column_names = "ground_truths"
        if (
            isinstance(ds.features[column_names], Sequence)
            and ds.features[column_names].feature.dtype == "string"
        ):
            logger.warning(
                "passing column names as 'ground_truths' is deprecated and will be removed in the next version, please use 'ground_truth' instead. Note that `ground_truth` should be of type string and not Sequence[string] like `ground_truths`"
            )
            gt = [gt[0] for gt in ds["ground_truths"]]
            ds = ds.add_column(
                "ground_truth",
                gt,
                new_fingerprint=ds._fingerprint
                + "a",  # adding random to fingerprint to avoid caching
            )
    return ds


def validate_column_dtypes(ds: Dataset):
    for column_names in ["question", "answer", "ground_truth"]:
        if column_names in ds.features:
            if ds.features[column_names].dtype != "string":
                raise ValueError(
                    f'Dataset feature "{column_names}" should be of type string'
                )

    for column_names in ["contexts"]:
        if column_names in ds.features:
            if not (
                isinstance(ds.features[column_names], Sequence)
                and ds.features[column_names].feature.dtype == "string"
            ):
                raise ValueError(
                    f'Dataset feature "{column_names}" should be of type'
                    f" Sequence[string], got {type(ds.features[column_names])}"
                )


EVALMODE_TO_COLUMNS = {
    EvaluationMode.qac: ["question", "answer", "contexts"],
    EvaluationMode.qa: ["question", "answer"],
    EvaluationMode.qc: ["question", "contexts"],
    EvaluationMode.gc: ["ground_truth", "contexts"],
    EvaluationMode.ga: ["ground_truth", "answer"],
    EvaluationMode.qga: ["question", "ground_truth", "answer"],
    EvaluationMode.qcg: ["question", "contexts", "ground_truth"],
    EvaluationMode.ca: ["contexts", "answer"],
}


def validate_evaluation_modes(
    ds: Dataset, metrics: list[Metric], evalmode_to_columns: dict = EVALMODE_TO_COLUMNS
):
    """
    validates the dataset and returns the evaluation type

    possible evaluation types
    1. (q,a,c)
    2. (q,a)
    3. (q,c)
    4. (g,a)
    """
    for m in metrics:
        required_columns = set(evalmode_to_columns[m.evaluation_mode])
        available_columns = set(ds.features.keys())
        if not required_columns.issubset(available_columns):
            extra_msg = ""
            if (
                isinstance(m, ContextPrecision)
                and "ground_truth" not in available_columns
            ):
                extra_msg = "Looks like you're trying to use 'context_precision' without ground_truth. Please use consider using  `context_utilization' instead."

            raise ValueError(
                f"The metric [{m.name}] that that is used requires the following "
                f"additional columns {list(required_columns - available_columns)} "
                f"to be present in the dataset. {extra_msg}"
            )



import logging
import typing as t
from dataclasses import dataclass

import numpy as np
from tenacity import (
    AsyncRetrying,
    Retrying,
    WrappedFn,
    after_log,
    retry_if_exception_type,
    stop_after_attempt,
    wait_random_exponential,
)
from tenacity.after import after_nothing


@dataclass
class RunConfig:
    """
    Configuration for a timeouts, retries and seed for Ragas operations.

    Parameters
    ----------
    timeout : int, optional
        Maximum time (in seconds) to wait for a single operation, by default 60.
    max_retries : int, optional
        Maximum number of retry attempts, by default 10.
    max_wait : int, optional
        Maximum wait time (in seconds) between retries, by default 60.
    max_workers : int, optional
        Maximum number of concurrent workers, by default 16.
    exception_types : Union[Type[BaseException], Tuple[Type[BaseException], ...]], optional
        Exception types to catch and retry on, by default (Exception,).
    log_tenacity : bool, optional
        Whether to log retry attempts using tenacity, by default False.
    seed : int, optional
        Random seed for reproducibility, by default 42.

    Attributes
    ----------
    rng : numpy.random.Generator
        Random number generator initialized with the specified seed.

    Notes
    -----
    The `__post_init__` method initializes the `rng` attribute as a numpy random
    number generator using the specified seed.
    """

    timeout: int = 180
    max_retries: int = 10
    max_wait: int = 60
    max_workers: int = 16
    exception_types: t.Union[
        t.Type[BaseException],
        t.Tuple[t.Type[BaseException], ...],
    ] = (Exception,)
    log_tenacity: bool = False
    seed: int = 42

    def __post_init__(self):
        self.rng = np.random.default_rng(seed=self.seed)


def add_retry(fn: WrappedFn, run_config: RunConfig) -> WrappedFn:
    # configure tenacity's after section wtih logger
    if run_config.log_tenacity is not None:
        logger = logging.getLogger(f"ragas.retry.{fn.__name__}")
        tenacity_logger = after_log(logger, logging.DEBUG)
    else:
        tenacity_logger = after_nothing

    r = Retrying(
        wait=wait_random_exponential(multiplier=1, max=run_config.max_wait),
        stop=stop_after_attempt(run_config.max_retries),
        retry=retry_if_exception_type(run_config.exception_types),
        reraise=True,
        after=tenacity_logger,
    )
    return r.wraps(fn)


def add_async_retry(fn: WrappedFn, run_config: RunConfig) -> WrappedFn:
    """
    Decorator for retrying a function if it fails.
    """
    # configure tenacity's after section wtih logger
    if run_config.log_tenacity is not None:
        logger = logging.getLogger(f"TENACITYRetry[{fn.__name__}]")
        tenacity_logger = after_log(logger, logging.DEBUG)
    else:
        tenacity_logger = after_nothing

    r = AsyncRetrying(
        wait=wait_random_exponential(multiplier=1, max=run_config.max_wait),
        stop=stop_after_attempt(run_config.max_retries),
        retry=retry_if_exception_type(run_config.exception_types),
        reraise=True,
        after=tenacity_logger,
    )
    return r.wraps(fn)



import re
import typing as t

import numpy as np
from langchain_core.documents import Document as LCDocument
from ragas_experimental.testset.graph import Node, NodeLevel, NodeType, Relationship
from ragas_experimental.testset.utils import merge_dicts


class HeadlineSplitter:
    def __init__(
        self,
        max_leaf_size: int = 4000,
        length_function=len,
        common_metadata_keys: t.List[str] = ["source"],
    ):
        self._length_function = length_function
        self._max_leaf_size = max_leaf_size
        self._common_metadata_keys = common_metadata_keys

    def _find_headline_indices(self, text, headlines):
        index_dict = {}
        for headline in headlines:
            # Build a regex pattern to match the headline with newlines before and after
            pattern = rf"(?<=\n){re.escape(headline)}"
            matches = re.finditer(pattern, text)
            first_match = next(matches, None)
            if first_match:
                index_dict[headline] = first_match.start()
        return index_dict

    def _split_text_by_headlines(self, text: str, headlines: t.List[str]):
        indices = []
        seperators = []
        headline_indices = self._find_headline_indices(text, headlines)
        values = list(headline_indices.values())
        keys = list(headline_indices.keys())
        values.append(len(text))
        for key, start_idx, end_idx in zip(keys, values[:-1], values[1:]):
            indices.append((start_idx, end_idx))
            seperators.append(key)

        chunks = [(text[idx[0] : idx[1]], sep) for idx, sep in zip(indices, seperators)]
        return chunks

    def _reassign_metadata(self, document: LCDocument, chunks: t.List[LCDocument]):
        extractive_metadata_keys = document.metadata.get("extractive_metadata_keys", [])
        for chunk in chunks:
            page_content = chunk.page_content
            text_chunk_metadata = {"extractive_metadata_keys": extractive_metadata_keys}
            for metadata_key in extractive_metadata_keys:
                metadata = document.metadata.get(metadata_key)
                if isinstance(metadata, str):
                    idx = page_content.find(metadata)
                    if idx != -1:
                        text_chunk_metadata[metadata_key] = metadata
                elif isinstance(metadata, list):
                    metadata_match_idx = [page_content.find(item) for item in metadata]
                    metadata_idx = [
                        idx
                        for idx, match_idx in enumerate(metadata_match_idx)
                        if match_idx != -1
                    ]
                    if metadata_idx:
                        text_chunk_metadata[metadata_key] = [
                            metadata[i] for i in metadata_idx
                        ]
            text_chunk_metadata = merge_dicts(chunk.metadata, text_chunk_metadata)
            chunk.metadata = text_chunk_metadata
            for key in self._common_metadata_keys:
                if key in document.metadata:
                    chunk.metadata[key] = document.metadata[key]
        return chunks

    def _split_document(
        self, document: LCDocument, headlines: t.List[str]
    ) -> t.List[LCDocument]:
        chunks = self._split_text_by_headlines(document.page_content, headlines)
        chunks = [
            LCDocument(page_content=chunk[0], metadata={"headline": chunk[1]})
            for chunk in chunks
        ]
        chunks = self._reassign_metadata(document, chunks)
        return chunks

    def _get_nodes_relationships(
        self,
        node: Node,
        headlines: t.List[str],
        nodes: t.List[Node],
        relationships: t.List[Relationship],
    ):
        document = LCDocument(
            page_content=node.properties["page_content"],
            metadata=node.properties["metadata"],
        )
        node_level = node.level
        chunks = self._split_document(document, headlines)
        node_idx = [idx for idx, n in enumerate(nodes) if n.id == node.id][0]
        for chunk in chunks:
            nodes.append(
                Node(
                    label=NodeType.CHUNK,
                    properties={
                        "page_content": chunk.page_content,
                        "metadata": chunk.metadata,
                    },
                    relationships=[],
                    level=node_level.next_level(),
                )
            )
            relationship = Relationship(
                source=node,
                target=nodes[-1],
                label="child",
                properties={"seperator": chunk.metadata["headline"]},
            )
            relationships.append(relationship)
            nodes[node_idx].relationships.append(relationship)
            relationship = Relationship(
                source=nodes[-1],
                target=node,
                label="parent",
                properties={"seperator": chunk.metadata["headline"]},
            )
            nodes[-1].relationships.append(relationship)
            relationships.append(relationship)
        return nodes, relationships

    def split_nodes_by_attribute(self, document: LCDocument, attribute: str):
        node = Node(
            label=NodeType.DOC,
            properties={
                "page_content": document.page_content,
                "metadata": document.metadata,
            },
            relationships=[],
            level=NodeLevel.LEVEL_0,
        )
        nodes = []
        relationships = []
        nodes.append(node)
        attr_values = document.metadata.get(attribute, [])
        if isinstance(attr_values, dict):
            headlines = attr_values.keys()
            nodes, relationships = self._get_nodes_relationships(
                node, headlines, nodes, relationships
            )
            level_1_nodes = [node for node in nodes if node.level == NodeLevel.LEVEL_1]
            for node in level_1_nodes:
                subheadings = attr_values.get(
                    node.properties["metadata"]["headline"], []
                )
                if subheadings:
                    nodes, relationships = self._get_nodes_relationships(
                        node, subheadings, nodes, relationships
                    )

        nodes, relationships = self._ensure_leaf_node_size(nodes, relationships)
        return nodes, relationships

    def get_all_leaf_nodes(self, nodes: t.List[Node]) -> t.List[Node]:
        leaf_nodes = []
        for node in nodes:
            child_relationships = [
                relation
                for relation in node.relationships
                if relation.label == "child" and relation.source.id == node.id
            ]
            if not child_relationships:
                leaf_nodes.append(node)

        return leaf_nodes

    def _ensure_leaf_node_size(
        self, nodes: t.List[Node], relationships: t.List[Relationship]
    ) -> t.Tuple[t.List[Node], t.List[Relationship]]:
        leaf_nodes = self.get_all_leaf_nodes(nodes)
        for leaf_node in leaf_nodes:
            chunks = []
            seperator = leaf_node.properties["metadata"].get("headline")
            doc = LCDocument(
                page_content=leaf_node.properties["page_content"],
                metadata=leaf_node.properties["metadata"],
            )
            page_content = leaf_node.properties["page_content"]
            content_length = self._length_function(page_content)
            if content_length > self._max_leaf_size:
                num_chunks = np.ceil(content_length / self._max_leaf_size)
                start_idx = [
                    max(0, i * self._max_leaf_size - 200)
                    for i in range(int(num_chunks))
                ]
                end_idx = [
                    (i + 1) * self._max_leaf_size for i in range(int(num_chunks))
                ]
                for start, end in zip(start_idx, end_idx):
                    chunks.append(
                        LCDocument(
                            page_content=page_content[start:end],
                            metadata={},
                        )
                    )
                chunks = self._reassign_metadata(doc, chunks)
                leaf_node_idx = [
                    idx for idx, n in enumerate(nodes) if n.id == leaf_node.id
                ][0]
                for chunk in chunks:
                    nodes.append(
                        Node(
                            label=NodeType.CHUNK,
                            properties={
                                "page_content": chunk.page_content,
                                "metadata": chunk.metadata,
                            },
                            relationships=[],
                            level=leaf_node.level.next_level(),
                        )
                    )
                    relationship = Relationship(
                        source=leaf_node,
                        target=nodes[-1],
                        label="child",
                        properties={"seperator": seperator},
                    )
                    relationships.append(relationship)
                    nodes[leaf_node_idx].relationships.append(relationship)
                    relationship = Relationship(
                        source=nodes[-1],
                        target=leaf_node,
                        label="parent",
                        properties={"seperator": seperator},
                    )
                    nodes[-1].relationships.append(relationship)
                    relationships.append(relationship)

        return nodes, relationships

    def split_documents(self, documents: t.Sequence[LCDocument], attribute: str):
        nodes = []
        relationships = []
        for doc in documents:
            doc_nodes, doc_relationships = self.split_nodes_by_attribute(doc, attribute)
            nodes.extend(doc_nodes)
            relationships.extend(doc_relationships)

        return nodes, relationships



import typing as t
from abc import ABC, abstractmethod
from dataclasses import dataclass, field

import pandas as pd
from langchain_core.documents import Document
from llama_index.core.schema import Document as LlamaindexDocument
from ragas_experimental.testset.questions import QAC

from ragas.embeddings import BaseRagasEmbeddings, embedding_factory
from ragas.llms.base import BaseRagasLLM, llm_factory


@dataclass
class QADistribution:
    question_types: t.List[t.Any]
    probabilities: t.List[float]

    def keys(self):
        return self.question_types

    def values(self):
        return self.probabilities

    def items(self):
        return zip(self.keys(), self.values())


@dataclass
class TestDataset:
    qac: t.List[QAC]

    def to_pandas(self):
        data = []
        for row in self.qac:
            data.append(row.to_dict())
        return pd.DataFrame(data)


@dataclass
class TestGenerator(ABC):
    llm: BaseRagasLLM = field(default_factory=llm_factory)
    embedding: BaseRagasEmbeddings = field(default_factory=embedding_factory)

    @abstractmethod
    def generate(
        self,
        docs: t.Sequence[Document],
        test_size: int,
        distribution: QADistribution,
    ) -> TestDataset:
        ...

    def generate_with_langchain_docs(
        self,
        docs: t.Sequence[Document],
        test_size: int,
        distribution: QADistribution,
    ) -> TestDataset:
        return self.generate(docs, test_size, distribution)

    def generate_with_llamaindex_docs(
        self,
        docs: t.Sequence[LlamaindexDocument],
        test_size: int,
        distribution: QADistribution,
    ) -> TestDataset:
        docs = [doc.to_langchain_format() for doc in docs]
        return self.generate(docs, test_size, distribution)



from __future__ import annotations

import logging
import typing as t
from abc import ABC
from dataclasses import dataclass, field

from ragas.run_config import RunConfig
from ragas.testset.prompts import (
    context_scoring_parser,
    context_scoring_prompt,
    evolution_elimination_parser,
    evolution_elimination_prompt,
    filter_question_prompt,
    question_filter_parser,
)

if t.TYPE_CHECKING:
    from ragas.llms.base import BaseRagasLLM
    from ragas.llms.prompt import Prompt
    from ragas.testset.docstore import Node


logger = logging.getLogger(__name__)


@dataclass
class Filter(ABC):
    llm: BaseRagasLLM

    def set_run_config(self, run_config: RunConfig):
        self.llm.set_run_config(run_config)

    def adapt(self, language: str, cache_dir: t.Optional[str] = None) -> None:
        """
        Adapt the filter to a different language.
        """
        raise NotImplementedError("adapt() is not implemented for {} Filter")

    def save(self, cache_dir: t.Optional[str] = None) -> None:
        """
        Save the filter prompts to a path.
        """
        raise NotImplementedError("save() is not implemented for {} Filter")


@dataclass
class NodeFilter(Filter):
    threshold: float = 1.5
    context_scoring_prompt: Prompt = field(
        default_factory=lambda: context_scoring_prompt
    )

    async def filter(self, node: Node) -> t.Dict:
        prompt = self.context_scoring_prompt.format(context=node.page_content)
        results = await self.llm.generate(prompt=prompt)
        output = results.generations[0][0].text.strip()
        output = await context_scoring_parser.aparse(output, prompt, self.llm)
        output = output.dict() if output is not None else {}
        output["score"] = sum(output.values()) / len(output.values())
        logger.debug("context scoring: %s", output)
        output.update({"score": output.get("score", 0) >= self.threshold})
        return output

    def adapt(self, language: str, cache_dir: t.Optional[str] = None) -> None:
        """
        Adapt the filter to a different language.
        """
        self.context_scoring_prompt = self.context_scoring_prompt.adapt(
            language, self.llm, cache_dir
        )

    def save(self, cache_dir: t.Optional[str] = None) -> None:
        """
        Save the filter prompts to a path.
        """
        self.context_scoring_prompt.save(cache_dir)


@dataclass
class QuestionFilter(Filter):
    llm: BaseRagasLLM
    filter_question_prompt: Prompt = field(
        default_factory=lambda: filter_question_prompt
    )

    async def filter(self, question: str) -> t.Tuple[bool, str]:
        prompt = self.filter_question_prompt.format(question=question)
        results = await self.llm.generate(prompt=prompt)
        results = results.generations[0][0].text.strip()
        results = await question_filter_parser.aparse(results, prompt, self.llm)
        results = results.dict() if results is not None else {}
        logger.debug("filtered question: %s", results)
        return results.get("verdict") == 1, results.get("feedback", "")

    def adapt(self, language: str, cache_dir: t.Optional[str] = None) -> None:
        """
        Adapt the filter to a different language.
        """
        self.filter_question_prompt = self.filter_question_prompt.adapt(
            language, self.llm, cache_dir
        )

    def save(self, cache_dir: t.Optional[str] = None) -> None:
        """
        Save the filter prompts to a path.
        """
        self.filter_question_prompt.save(cache_dir)


@dataclass
class EvolutionFilter(Filter):
    llm: BaseRagasLLM
    evolution_elimination_prompt: Prompt = field(
        default_factory=lambda: evolution_elimination_prompt
    )

    async def filter(self, simple_question: str, compressed_question: str) -> bool:
        prompt = self.evolution_elimination_prompt.format(
            question1=simple_question, question2=compressed_question
        )
        results = await self.llm.generate(prompt=prompt)
        results = results.generations[0][0].text.strip()
        results = await evolution_elimination_parser.aparse(results, prompt, self.llm)
        results = results.dict() if results is not None else {}
        logger.debug("evolution filter: %s", results)
        return results.get("verdict") == 1

    def adapt(self, language: str, cache_dir: t.Optional[str] = None) -> None:
        """
        Adapt the filter to a different language.
        """
        self.evolution_elimination_prompt = self.evolution_elimination_prompt.adapt(
            language, self.llm, cache_dir
        )

    def save(self, cache_dir: t.Optional[str] = None) -> None:
        """
        Save the filter prompts to a path.
        """
        self.evolution_elimination_prompt.save(cache_dir)



from ragas_experimental.testset.relationships.base import Similarity
from ragas_experimental.testset.relationships.relation import RelationshipBuilder
from ragas_experimental.testset.relationships.similarity import Cosine, Jaccard

__all__ = ["Similarity", "RelationshipBuilder", "Jaccard", "Cosine"]



import typing as t
from abc import ABC, abstractmethod
from dataclasses import dataclass

from ragas_experimental.testset.graph import Node


@dataclass
class Similarity(ABC):
    name: str
    attribute1: str
    attribute2: str

    def get_attribute(self, node: Node, attribute: str):
        if attribute == "page_content":
            return node.properties["page_content"]
        elif attribute in node.properties["metadata"]:
            return node.properties["metadata"][attribute]
        else:
            raise ValueError(f"Attribute {attribute} not found in node")

    @abstractmethod
    def extract(self, x_nodes: t.List[Node], y_nodes: t.List[Node]) -> t.Any:
        pass



from __future__ import annotations

import re
import warnings

import numpy as np

rng = np.random.default_rng(seed=42)


def load_as_score(text):
    """
    validate and returns given text as score
    """

    pattern = r"^[\d.]+$"
    if not re.match(pattern, text):
        warnings.warn("Invalid score")
        score = 0.0
    else:
        score = eval(text)

    return score



import pytest
from datasets import load_dataset

from ragas import evaluate
from ragas.metrics import (
    answer_relevancy,
    context_precision,
    context_recall,
    faithfulness,
)

# loading the V2 dataset
amnesty_qa = load_dataset("explodinggradients/amnesty_qa", "english_v2")["eval"]


def assert_in_range(score: float, value: float, plus_or_minus: float):
    """
    Check if computed score is within the range of value +/- max_range
    """
    assert value - plus_or_minus <= score <= value + plus_or_minus


@pytest.mark.ragas_ci
def test_amnesty_e2e():
    result = evaluate(
        amnesty_qa,
        metrics=[answer_relevancy, faithfulness, context_recall, context_precision],
        in_ci=True,
    )
    assert result["answer_relevancy"] >= 0.9
    assert result["context_recall"] >= 0.95
    assert result["context_precision"] >= 0.95
    assert_in_range(result["faithfulness"], value=0.4, plus_or_minus=0.1)


@pytest.mark.ragas_ci
def test_assert_in_range():
    assert_in_range(0.5, value=0.1, plus_or_minus=0.1)



import typing as t
from dataclasses import dataclass

import numpy as np
from ragas_experimental.testset.graph import Node
from ragas_experimental.testset.relationships.base import Similarity


@dataclass
class Jaccard(Similarity):
    type: t.Optional[str] = None
    threshold: t.Optional[int] = 80

    def __post_init__(self):
        if self.type == "fuzzy":
            try:
                from fuzzywuzzy import fuzz
            except ImportError:
                raise ImportError(
                    "fuzzywuzzy is not installed. Run pip install fuzzywuzzy"
                )
            self.fuzz = fuzz
            self.threshold = self.threshold or 80

    def _calculate_fuzz(self, x: t.List[str], y: t.List[str]):
        fuzz_scores = 0
        for item in x:
            fuzz_scores = 0
            for element in y:
                if self.fuzz.ratio(item, element) >= (self.threshold or 0):
                    fuzz_scores += 1
        return fuzz_scores

    def extract(self, x_nodes: t.List[Node], y_nodes: t.List[Node]):
        jaccard_similarity_matrix = np.zeros((len(x_nodes), len(y_nodes)))

        doc1_items = [self.get_attribute(doc, self.attribute1) for doc in x_nodes]
        doc2_items = [self.get_attribute(doc, self.attribute2) for doc in y_nodes]
        if all(isinstance(item, list) for item in doc1_items) and all(
            isinstance(item, list) for item in doc2_items
        ):
            for i, a in enumerate(doc1_items):
                for k, b in enumerate(doc2_items):
                    if self.type == "fuzzy":
                        intersection = self._calculate_fuzz(a, b)
                    else:
                        intersection = len(set(a).intersection(set(b)))
                    union = len(set(a).union(set(b)))
                    jaccard_similarity_matrix[i][k] = (
                        intersection / union if union != 0 else 0
                    )
                    jaccard_similarity_matrix[i][k] = intersection / union

        return jaccard_similarity_matrix


@dataclass
class Cosine(Similarity):
    def extract(self, x_nodes: t.List[Node], y_nodes: t.List[Node]) -> t.Any:
        embeddings_1 = [self.get_attribute(doc, self.attribute1) for doc in x_nodes]
        embeddings_2 = [self.get_attribute(doc, self.attribute2) for doc in y_nodes]
        embeddings_1 = np.array(embeddings_1)
        embeddings_2 = np.array(embeddings_2)
        cosine_similarity_matrix = np.dot(embeddings_1, embeddings_2.T) / (
            np.linalg.norm(embeddings_1, axis=1) * np.linalg.norm(embeddings_2, axis=1)
        )
        return cosine_similarity_matrix



import asyncio
import logging
import os
import subprocess
from pathlib import Path
from typing import Any, AsyncIterator, Dict, Iterator, List, Union

import aiofiles
from langchain_community.document_loaders.helpers import detect_file_encodings
from langchain_community.document_loaders.unstructured import UnstructuredBaseLoader
from langchain_core.documents import Document

# Setup logging
logger = logging.getLogger(__name__)


class RAGASLoader(UnstructuredBaseLoader):
    def __init__(
        self,
        file_path: Union[str, Path, List[Union[str, Path]]],
        mode: str = "single",
        autodetect_encoding: bool = False,
        **unstructured_kwargs: Any,
    ):
        super().__init__(mode=mode, **unstructured_kwargs)
        self.autodetect_encoding = autodetect_encoding
        if isinstance(file_path, list):
            self.file_paths = [Path(fp) for fp in file_path]
        else:
            self.file_paths = [Path(file_path)]

    def _get_elements(self, file_path: Path) -> List:
        from unstructured.partition.auto import partition

        try:
            return partition(filename=str(file_path), **self.unstructured_kwargs)
        except Exception as e:
            logger.error("Error in _get_elements for file %s: %s", file_path, e)
            return []

    def _get_metadata(self, file_path: Path) -> Dict:
        try:
            return {"source": str(file_path), "raw_content": self._read_file(file_path)}
        except Exception as e:
            logger.error("Error in _get_metadata for file %s: %s", file_path, e)
            return {"source": str(file_path), "raw_content": ""}

    async def _aget_metadata(self, file_path: Path) -> Dict:
        try:
            return {
                "source": str(file_path),
                "raw_content": await self._aread_file(file_path),
            }
        except Exception as e:
            logger.error("Error in _aget_metadata for file %s: %s", file_path, e)
            return {"source": str(file_path), "raw_content": ""}

    def _read_file(self, file_path: Path) -> str:
        try:
            with open(file_path, "r", encoding="utf-8") as f:
                return f.read()
        except UnicodeDecodeError as e:
            if self.autodetect_encoding:
                detected_encodings = detect_file_encodings(file_path)
                for encoding in detected_encodings:
                    try:
                        with file_path.open(encoding=encoding.encoding) as f:
                            return f.read()
                    except UnicodeDecodeError:
                        continue
                logger.error("Failed to decode %s with detected encodings.", file_path)
                raise RuntimeError(
                    f"Failed to decode {file_path} with detected encodings."
                )
            else:
                logger.error(
                    "Error loading %s due to encoding issue.", file_path, exc_info=e
                )
                raise RuntimeError(
                    f"Error loading {file_path} due to encoding issue."
                ) from e
        except Exception as e:
            logger.error(
                "Error loading %s due to an unexpected error: %s", file_path, e
            )
            raise RuntimeError(
                f"Error loading {file_path} due to an unexpected error: {e}"
            ) from e

    async def _aread_file(self, file_path: Path) -> str:
        try:
            async with aiofiles.open(file_path, mode="r", encoding="utf-8") as f:
                return await f.read()
        except UnicodeDecodeError as e:
            if self.autodetect_encoding:
                detected_encodings = await asyncio.to_thread(
                    detect_file_encodings, file_path
                )
                for encoding in detected_encodings:
                    try:
                        async with aiofiles.open(
                            file_path, mode="r", encoding=encoding.encoding
                        ) as f:
                            return await f.read()
                    except UnicodeDecodeError:
                        continue
                logger.error("Failed to decode %s with detected encodings.", file_path)
                raise RuntimeError(
                    f"Failed to decode {file_path} with detected encodings."
                )
            else:
                logger.error(
                    "Error loading %s due to encoding issue.", file_path, exc_info=e
                )
                raise RuntimeError(
                    f"Error loading {file_path} due to encoding issue."
                ) from e
        except Exception as e:
            logger.error(
                "Error loading %s due to an unexpected error: %s", file_path, e
            )
            raise RuntimeError(
                f"Error loading {file_path} due to an unexpected error: {e}"
            ) from e

    def _load_directory(self, directory: Path) -> Iterator[Document]:
        file_extensions = ["xml", "md", "txt", "html", "ppt", "ppx", "pdf"]
        for root, _, files in os.walk(directory):
            for file in files:
                if any(file.endswith(ext) for ext in file_extensions):
                    file_path = Path(root) / file
                    if file.endswith("pdf"):
                        output_dir = Path("experimental_notebook/markdown")
                        output_file = (
                            output_dir / file_path.stem / file.replace(".pdf", ".md")
                        )
                        command = [
                            "marker_single",
                            str(file_path),
                            str(output_dir),
                            "--batch_multiplier",
                            "2",
                            "--max_pages",
                            "10",
                            "--langs",
                            "English",
                        ]
                        try:
                            subprocess.run(
                                command, check=True, capture_output=True, text=True
                            )
                            logger.info("Processed %s to %s", file_path, output_file)
                            file_path = output_file
                        except subprocess.CalledProcessError as e:
                            logger.error(
                                "An error occurred while processing %s:", file_path
                            )
                            logger.error(e.stderr)
                            continue
                        except FileNotFoundError:
                            logger.error(
                                "The 'marker_single' command was not found. Make sure it's installed and in your PATH."
                            )
                            continue

                    if file_path.is_file():
                        yield from self._load_file(file_path)

    async def _aload_directory(self, directory: Path) -> AsyncIterator[Document]:
        file_extensions = ["xml", "md", "txt", "html", "ppt", "ppx", "pdf"]
        for root, _, files in os.walk(directory):
            for file in files:
                if any(file.endswith(ext) for ext in file_extensions):
                    file_path = Path(root) / file
                    if file.endswith("pdf"):
                        output_dir = Path("experimental_notebook/markdown")
                        output_file = (
                            output_dir / file_path.stem / file.replace(".pdf", ".md")
                        )
                        command = [
                            "marker_single",
                            str(file_path),
                            str(output_dir),
                            "--batch_multiplier",
                            "2",
                            "--max_pages",
                            "10",
                            "--langs",
                            "English",
                        ]
                        try:
                            subprocess.run(
                                command, check=True, capture_output=True, text=True
                            )
                            logger.info("Processed %s to %s", file_path, output_file)
                            file_path = output_file
                        except subprocess.CalledProcessError as e:
                            logger.error(
                                "An error occurred while processing %s:", file_path
                            )
                            logger.error(e.stderr)
                            continue
                        except FileNotFoundError:
                            logger.error(
                                "The 'marker_single' command was not found. Make sure it's installed and in your PATH."
                            )
                            continue

                    if file_path.is_file():
                        async for document in self._aload_file(file_path):
                            yield document

    def lazy_load(self) -> Iterator[Document]:
        try:
            for file_path in self.file_paths:
                if file_path.is_dir():
                    yield from self._load_directory(file_path)
                elif file_path.is_file():
                    yield from self._load_file(file_path)
                else:
                    logger.error(
                        "The path %s does not exist or is neither a directory nor a file.",
                        file_path,
                    )
                    raise ValueError(
                        f"The path {file_path} does not exist or is neither a directory nor a file."
                    )
        except Exception as e:
            logger.error("Error loading file or directory: %s", e)
            raise RuntimeError(f"Error loading file or directory: {e}")

    async def lazy_aload(self) -> AsyncIterator[Document]:
        try:
            for file_path in self.file_paths:
                if file_path.is_dir():
                    async for document in self._aload_directory(file_path):
                        yield document
                elif file_path.is_file():
                    async for document in self._aload_file(file_path):
                        yield document
                else:
                    logger.error(
                        "The path %s does not exist or is neither a directory nor a file.",
                        file_path,
                    )
                    raise ValueError(
                        f"The path {file_path} does not exist or is neither a directory nor a file."
                    )
        except Exception as e:
            logger.error("Error loading file or directory: %s", e)
            raise RuntimeError(f"Error loading file or directory: {e}")

    def _load_file(self, file_path: Path) -> Iterator[Document]:
        """Load file."""
        elements = self._get_elements(file_path)
        self._post_process_elements(elements)
        if self.mode == "elements":
            for element in elements:
                metadata = self._get_metadata(file_path)
                if hasattr(element, "metadata"):
                    metadata.update(element.metadata.to_dict())
                if hasattr(element, "category"):
                    metadata["category"] = element.category
                yield Document(page_content=str(element), metadata=metadata)
        elif self.mode == "paged":
            text_dict: Dict[int, str] = {}
            meta_dict: Dict[int, Dict] = {}

            for idx, element in enumerate(elements):
                metadata = self._get_metadata(file_path)
                if hasattr(element, "metadata"):
                    metadata.update(element.metadata.to_dict())
                page_number = metadata.get("page_number", 1)

                # Check if this page_number already exists in docs_dict
                if page_number not in text_dict:
                    # If not, create new entry with initial text and metadata
                    text_dict[page_number] = str(element) + "\n\n"
                    meta_dict[page_number] = metadata
                else:
                    # If exists, append to text and update the metadata
                    text_dict[page_number] += str(element) + "\n\n"
                    meta_dict[page_number].update(metadata)

            # Convert the dict to a list of Document objects
            for key in text_dict.keys():
                yield Document(page_content=text_dict[key], metadata=meta_dict[key])
        elif self.mode == "single":
            metadata = self._get_metadata(file_path)
            text = "\n\n".join([str(el) for el in elements])
            yield Document(page_content=text, metadata=metadata)
        else:
            logger.error("Mode %s not supported.", self.mode)
            raise ValueError(f"mode of {self.mode} not supported.")

    async def _aload_file(self, file_path: Path) -> AsyncIterator[Document]:
        elements = await asyncio.to_thread(self._get_elements, file_path)
        self._post_process_elements(elements)
        metadata = await self._aget_metadata(file_path)
        if self.mode == "elements":
            for element in elements:
                element_metadata = metadata.copy()
                if hasattr(element, "metadata"):
                    element_metadata.update(element.metadata.to_dict())
                if hasattr(element, "category"):
                    element_metadata["category"] = element.category
                yield Document(page_content=str(element), metadata=element_metadata)
        elif self.mode == "paged":
            text_dict = {}
            meta_dict = {}
            for idx, element in enumerate(elements):
                element_metadata = metadata.copy()
                if hasattr(element, "metadata"):
                    element_metadata.update(element.metadata.to_dict())
                page_number = element_metadata.get("page_number", 1)

                if page_number not in text_dict:
                    text_dict[page_number] = str(element) + "\n\n"
                    meta_dict[page_number] = element_metadata
                else:
                    text_dict[page_number] += str(element) + "\n\n"
                    meta_dict[page_number].update(element_metadata)

            for key in text_dict.keys():
                yield Document(page_content=text_dict[key], metadata=meta_dict[key])
        elif self.mode == "single":
            text = "\n\n".join([str(el) for el in elements])
            yield Document(page_content=text, metadata=metadata)
        else:
            logger.error("Mode %s not supported.", self.mode)
            raise ValueError(f"mode of {self.mode} not supported.")



from __future__ import annotations

import logging
import typing as t
from dataclasses import dataclass
from random import choices, sample

import pandas as pd
from datasets import Dataset
from langchain_openai.chat_models import ChatOpenAI
from langchain_openai.embeddings import OpenAIEmbeddings

from ragas._analytics import TestsetGenerationEvent, track, track_was_completed
from ragas.embeddings.base import (
    BaseRagasEmbeddings,
    LangchainEmbeddingsWrapper,
    LlamaIndexEmbeddingsWrapper,
)
from ragas.exceptions import ExceptionInRunner
from ragas.executor import Executor
from ragas.llms import BaseRagasLLM, LangchainLLMWrapper, LlamaIndexLLMWrapper
from ragas.run_config import RunConfig
from ragas.testset.docstore import Document, DocumentStore, InMemoryDocumentStore
from ragas.testset.evolutions import (
    ComplexEvolution,
    CurrentNodes,
    DataRow,
    Evolution,
    multi_context,
    reasoning,
    simple,
)
from ragas.testset.extractor import KeyphraseExtractor
from ragas.testset.filters import EvolutionFilter, NodeFilter, QuestionFilter
from ragas.utils import check_if_sum_is_close, deprecated, get_feature_language, is_nan

if t.TYPE_CHECKING:
    from langchain_core.documents import Document as LCDocument
    from langchain_core.embeddings import Embeddings as LangchainEmbeddings
    from langchain_core.language_models import BaseLanguageModel as LangchainLLM
    from llama_index.core.base.embeddings.base import (
        BaseEmbedding as LlamaIndexEmbeddings,
    )
    from llama_index.core.base.llms.base import BaseLLM as LlamaindexLLM
    from llama_index.core.schema import Document as LlamaindexDocument

logger = logging.getLogger(__name__)

Distributions = t.Dict[t.Any, float]
DEFAULT_DISTRIBUTION = {simple: 0.5, reasoning: 0.25, multi_context: 0.25}


@dataclass
class TestDataset:
    """
    TestDataset class
    """

    test_data: t.List[DataRow]

    def _to_records(self) -> t.List[t.Dict]:
        data_samples = []
        for data in self.test_data:
            data_dict = dict(data)
            data_dict["episode_done"] = True
            data_samples.append(data_dict)
        return data_samples

    def to_pandas(self) -> pd.DataFrame:
        return pd.DataFrame.from_records(self._to_records())

    def to_dataset(self) -> Dataset:
        return Dataset.from_list(self._to_records())


@dataclass
class TestsetGenerator:
    generator_llm: BaseRagasLLM
    critic_llm: BaseRagasLLM
    embeddings: BaseRagasEmbeddings
    docstore: DocumentStore

    @classmethod
    def from_langchain(
        cls,
        generator_llm: LangchainLLM,
        critic_llm: LangchainLLM,
        embeddings: LangchainEmbeddings,
        docstore: t.Optional[DocumentStore] = None,
        run_config: RunConfig = RunConfig(),
        chunk_size: int = 1024,
    ) -> "TestsetGenerator":
        generator_llm_model = LangchainLLMWrapper(generator_llm)
        critic_llm_model = LangchainLLMWrapper(critic_llm)
        embeddings_model = LangchainEmbeddingsWrapper(embeddings)

        keyphrase_extractor = KeyphraseExtractor(llm=generator_llm_model)
        if docstore is None:
            from langchain.text_splitter import TokenTextSplitter

            splitter = TokenTextSplitter(chunk_size=chunk_size, chunk_overlap=0)
            docstore = InMemoryDocumentStore(
                splitter=splitter,
                embeddings=embeddings_model,
                extractor=keyphrase_extractor,
                run_config=run_config,
            )
        return cls(
            generator_llm=generator_llm_model,
            critic_llm=critic_llm_model,
            embeddings=embeddings_model,
            docstore=docstore,
        )

    @classmethod
    def from_llama_index(
        cls,
        generator_llm: LlamaindexLLM,
        critic_llm: LlamaindexLLM,
        embeddings: LlamaIndexEmbeddings,
        docstore: t.Optional[DocumentStore] = None,
        run_config: RunConfig = RunConfig(),
    ) -> "TestsetGenerator":
        generator_llm_model = LlamaIndexLLMWrapper(generator_llm)
        critic_llm_model = LlamaIndexLLMWrapper(critic_llm)
        embeddings_model = LlamaIndexEmbeddingsWrapper(embeddings)
        keyphrase_extractor = KeyphraseExtractor(llm=generator_llm_model)
        if docstore is None:
            from langchain.text_splitter import TokenTextSplitter

            splitter = TokenTextSplitter(chunk_size=1024, chunk_overlap=0)
            docstore = InMemoryDocumentStore(
                splitter=splitter,
                embeddings=embeddings_model,
                extractor=keyphrase_extractor,
                run_config=run_config,
            )
        return cls(
            generator_llm=generator_llm_model,
            critic_llm=critic_llm_model,
            embeddings=embeddings_model,
            docstore=docstore,
        )

    @classmethod
    @deprecated("0.1.4", removal="0.2.0", alternative="from_langchain")
    def with_openai(
        cls,
        generator_llm: str = "gpt-4o-mini",
        critic_llm: str = "gpt-4o",
        embeddings: str = "text-embedding-ada-002",
        docstore: t.Optional[DocumentStore] = None,
        chunk_size: int = 1024,
    ) -> "TestsetGenerator":
        generator_llm_model = ChatOpenAI(model=generator_llm)
        critic_llm_model = ChatOpenAI(model=critic_llm)
        embeddings_model = OpenAIEmbeddings(model=embeddings)

        return cls.from_langchain(
            generator_llm=generator_llm_model,
            critic_llm=critic_llm_model,
            embeddings=embeddings_model,
            docstore=docstore,
            chunk_size=chunk_size,
        )

    def generate_with_llamaindex_docs(
        self,
        documents: t.Sequence[LlamaindexDocument],
        test_size: int,
        distributions: t.Optional[Distributions] = None,
        with_debugging_logs=False,
        is_async: bool = True,
        raise_exceptions: bool = True,
        run_config: t.Optional[RunConfig] = None,
    ):
        distributions = distributions or {}
        # chunk documents and add to docstore
        self.docstore.add_documents(
            [Document.from_llamaindex_document(doc) for doc in documents]
        )

        return self.generate(
            test_size=test_size,
            distributions=distributions,
            with_debugging_logs=with_debugging_logs,
            is_async=is_async,
            run_config=run_config,
            raise_exceptions=raise_exceptions,
        )

    # if you add any arguments to this function, make sure to add them to
    # generate_with_langchain_docs as well
    def generate_with_langchain_docs(
        self,
        documents: t.Sequence[LCDocument],
        test_size: int,
        distributions: t.Optional[Distributions] = None,
        with_debugging_logs=False,
        is_async: bool = True,
        raise_exceptions: bool = True,
        run_config: t.Optional[RunConfig] = None,
    ):
        distributions = distributions or {}
        # chunk documents and add to docstore
        self.docstore.add_documents(
            [Document.from_langchain_document(doc) for doc in documents]
        )

        return self.generate(
            test_size=test_size,
            distributions=distributions,
            with_debugging_logs=with_debugging_logs,
            is_async=is_async,
            raise_exceptions=raise_exceptions,
            run_config=run_config,
        )

    def init_evolution(self, evolution: Evolution) -> None:
        evolution.docstore = self.docstore

        if evolution.generator_llm is None:
            evolution.generator_llm = self.generator_llm

            if evolution.question_filter is None:
                evolution.question_filter = QuestionFilter(llm=self.critic_llm)
            if evolution.node_filter is None:
                evolution.node_filter = NodeFilter(llm=self.critic_llm)

            if isinstance(evolution, ComplexEvolution):
                if evolution.evolution_filter is None:
                    evolution.evolution_filter = EvolutionFilter(llm=self.critic_llm)

    @track_was_completed
    def generate(
        self,
        test_size: int,
        distributions: t.Optional[Distributions] = None,
        with_debugging_logs=False,
        is_async: bool = True,
        raise_exceptions: bool = True,
        run_config: t.Optional[RunConfig] = None,
    ):
        distributions = distributions or DEFAULT_DISTRIBUTION
        # validate distributions
        if not check_if_sum_is_close(list(distributions.values()), 1.0, 3):
            raise ValueError(
                f"distributions passed do not sum to 1.0 [got {sum(list(distributions.values()))}]. Please check the "
                f"distributions."
            )

        # configure run_config for docstore
        if run_config is None:
            run_config = RunConfig(max_retries=15, max_wait=90)
        self.docstore.set_run_config(run_config)

        # init filters and evolutions
        for evolution in distributions:
            self.init_evolution(evolution)
            evolution.init(is_async=is_async, run_config=run_config)

        if with_debugging_logs:
            from ragas.utils import patch_logger

            patch_logger("ragas.testset.evolutions", logging.DEBUG)
            patch_logger("ragas.testset.extractor", logging.DEBUG)
            patch_logger("ragas.testset.filters", logging.DEBUG)
            patch_logger("ragas.testset.docstore", logging.DEBUG)
            patch_logger("ragas.llms.prompt", logging.DEBUG)

        exec = Executor(
            desc="Generating",
            keep_progress_bar=True,
            raise_exceptions=raise_exceptions,
            run_config=run_config,
        )

        current_nodes = [
            CurrentNodes(root_node=n, nodes=[n])
            for n in self.docstore.get_random_nodes(k=test_size)
        ]
        total_evolutions = 0
        for evolution, probability in distributions.items():
            for i in sample(range(test_size), round(probability * test_size)):
                exec.submit(
                    evolution.evolve,
                    current_nodes[i],
                    name=f"{evolution.__class__.__name__}-{i}",
                )
                total_evolutions += 1
        if total_evolutions <= test_size:
            filler_evolutions = choices(
                list(distributions), k=test_size - total_evolutions
            )
            for evolution in filler_evolutions:
                exec.submit(
                    evolution.evolve,
                    current_nodes[total_evolutions],
                    name=f"{evolution.__class__.__name__}-{total_evolutions}",
                )
                total_evolutions += 1

        try:
            test_data_rows = exec.results()
            if not test_data_rows:
                raise ExceptionInRunner()

        except ValueError as e:
            raise e
        # make sure to ignore any NaNs that might have been returned
        # due to failed evolutions. MaxRetriesExceeded is a common reason
        test_data_rows = [r for r in test_data_rows if not is_nan(r)]
        test_dataset = TestDataset(test_data=test_data_rows)
        evol_lang = [get_feature_language(e) for e in distributions]
        evol_lang = [e for e in evol_lang if e is not None]
        track(
            TestsetGenerationEvent(
                event_type="testset_generation",
                evolution_names=[e.__class__.__name__.lower() for e in distributions],
                evolution_percentages=[distributions[e] for e in distributions],
                num_rows=len(test_dataset.test_data),
                language=evol_lang[0] if len(evol_lang) > 0 else "",
                is_experiment=False,
            )
        )

        return test_dataset

    def adapt(
        self,
        language: str,
        evolutions: t.List[Evolution],
        cache_dir: t.Optional[str] = None,
    ) -> None:
        assert isinstance(
            self.docstore, InMemoryDocumentStore
        ), "Must be an instance of in-memory docstore"
        assert self.docstore.extractor is not None, "Extractor is not set"

        self.docstore.extractor.adapt(language, cache_dir=cache_dir)
        for evolution in evolutions:
            self.init_evolution(evolution)
            evolution.init()
            evolution.adapt(language, cache_dir=cache_dir)

    def save(
        self, evolutions: t.List[Evolution], cache_dir: t.Optional[str] = None
    ) -> None:
        """
        Save the docstore prompts to a path.
        """
        assert isinstance(
            self.docstore, InMemoryDocumentStore
        ), "Must be an instance of in-memory docstore"
        assert self.docstore.extractor is not None, "Extractor is not set"

        self.docstore.extractor.save(cache_dir)
        for evolution in evolutions:
            assert evolution.node_filter is not None, "NodeFilter is not set"
            assert evolution.question_filter is not None, "QuestionFilter is not set"
            if isinstance(evolution, ComplexEvolution):
                assert (
                    evolution.evolution_filter is not None
                ), "EvolutionFilter is not set"
            evolution.save(cache_dir=cache_dir)
